# In order to understand how the code works, it is a good idea to check the
# final section of the file that starts with
#   if __name__ == '__main__'
#
# Your task is essentially to replace all the parts marked as TODO or as
# instructed through comments in the functions, as well as the filenames
# and labels in the main part of the code which can be found at the end of
# this file.
#
# The raise command is used to help you out in finding where you still need to
# write your own code. When you successfully modified the code in that part,
# remove the `raise` command.
import networkx as nx
import numpy as np
from numpy import random
import pandas as pd
import pickle
import time
import os

from matplotlib import pyplot as plt
from matplotlib import rc
rc.usetex = True
from matplotlib.animation import FuncAnimation
import scipy.stats

# Select data directory
if os.path.isdir('/coursedata/project'):
    project_data_dir = '/coursedata/project'
elif os.path.isdir('../data/project'):
    project_data_dir = '../data/project'
else:
    # Specify course_data_dir on your machine
    project_data_dir = 'some_path'
    # YOUR CODE HERE

event_file_path = project_data_dir + '/events_US_air_traffic_GMT.txt'
network_file_path = project_data_dir + '/aggregated_US_air_traffic_network_undir.edg'
airport_info_file_path = project_data_dir + '/US_airport_id_info.csv'
background_image_file_path = project_data_dir + '/US_air_bg.png'

# Visualization tools

# The following visualization tools are provided to help you. There is no need to modify these parts.

# The frunction plot_prevalence is used in multiple tasks from Task 2 onwards and plot_scatters_in_subplots 
# from Task 4 onwards. The function plot_network_USA is used in Task 6.

# As an extra visualization, you can find a function called visualize_si which animates the air transport using 
# matplotlib. The animations use the background image US_air_bg.png, the airport id-info file 
# US_airport_id_info.csv and the events file events_US_air_traffic_GMT.txt, all of which are available at 
# in the data zip. Note that this visualization is not needed for completion of the tasks, it’s just offered for 
# your benefit.

def plot_prevalence(prevalence_data, time_bins, colors, labels):
    """
    Plots the prevalence of the disease (proportion of infected
    nodes) as a function of time.
    
    Parameters:
    -----------
    prevalence_data : np.array
        the data to be plotted. Each row of prevalence_data is
        plotted as a separate plot. Number of columns should
        match the length of time_bins
    time_bins : np.array
        time bins used to calculate the prevalence. prevalence_data[N][i]
        should contain the proportion of nodes infected before time_bins[i]
        (in condition N)
    colors : list
        colors of the lines to be plotted. len(colors) should match the
        number of rows in prevalence_data
    labels : list
        labels of the lines to be plotted. len(labels) should match the
        number of rows in prevalence_data
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)
    time_bins = time_bins - time_bins[0] # origin shift: all times are now expressed as "after time_bins[0]"
    
    for prevalence, color, label in zip(prevalence_data, colors, labels):
        ax.plot(time_bins, prevalence, color = color, label = label)

    ax.legend(loc="best")
    ax.set_xlabel("Time in seconds after first event.")
    ax.grid()
    ax.set_ylabel("Prevalence of the disease")
    
def plot_scatters_in_subplots(x_vals, y_vals, labels, n_columns=2, y_label='Infection time', x_labels=[]):
    """
    Creates a set of scatter plots, each of them located in
    a subplot of its own.
    
    Parameters:
    -----------
    x_vals : list
        x values of the scatter plots (e.g. values of network properties to be plotted); y_vals[i]
        should be a list of y values of type i for all nodes
    y_vals: iterable
        y values of the scatter plots (e.g. median infection times); y_vals will be
        separately plotted against all elements of x_vals
    labels : list
        labels of the scatter points to be plotted. len(labels) should match the
        number of elements in x_vals.
    n_columns : int
        how many columns should the final figure contain, i.e. how many
        subplots there should be at each row
    y_label : str
        label of the y axis
    x_labels : list
        labels of the x axes for each subplot (given as strings); 
        if given, len(x_labels) should match the number of elements
        in x_vals.
    """
    fig = plt.figure()
    
    n_subplots = len(x_vals)
    n_rows = n_subplots / np.float(np.ceil(n_columns))
    
    if len(x_labels) == 0:
        x_labels = labels
    
    for i, (x_val, label, x_label) in enumerate(zip(x_vals, labels, x_labels)):
        ax = fig.add_subplot(n_rows, n_columns, i + 1)
        ax.scatter(x_val, y_vals, label=label, alpha=0.2)
        ax.set_ylabel(y_label)
        ax.set_xlabel(x_label)
        ax.legend(loc="best")
    fig.tight_layout()
    
def plot_network_usa(net, xycoords, bg_figname, edges=None, linewidths=None):
    """
    Plots the network on top of the USA map.
    This code makes use of the file US_air_bg.png available at /coursedata.
    If you want to use this function outside of Jupyterhub, the file
    should be located in the same directory where you run the code.

    Parameters
    ----------
    net : nx.Graph()
        the network to be plotted
    xycoords : dict
        dictionary of node_id to coordinates (x,y)
    bg_figname: str
        path of the background image used in the visualization (string); 
        in this project, the path of US_air_bg.png
    edges : list 
        a list of node index tuples (node_i,node_j),
        if None all network edges are plotted.
    linewidths : list
        a list with equal length and order to egdes -list.
        See nx.draw_networkx documentation

    example code:

    import random
    import pandas as pd
    import numpy as np
    import networkx as nx

    id_data = np.genfromtxt(csv_path, delimiter=',', dtype=None, names=True)
    xycoords = {}
    for row in id_data:
        xycoords[str(row['id'])] = (row['xcoordviz'], row['ycoordviz'])
    net = nx.read_weighted_edgelist(network_path)

    lw_list =[]
    edgelist =[]
    for edge in net.edges():
        lw_list.append(random.random()) # change this at least
        edgelist.append(edge) # edgelist created to maintain the right order
    
    plot_network_usa(net, xycoords, bg_fname, edges=edgelist, linewidths=lw_list)

    """
    fig = plt.figure()
    ax = fig.add_axes([0, 0, 1, 0.9])
    img = plt.imread(bg_figname)
    axis_extent = (-6674391.856090588, 4922626.076444283,
                   -2028869.260519173, 4658558.416671531)
    ax.imshow(img, extent=axis_extent)
    ax.set_xlim((axis_extent[0], axis_extent[1]))
    ax.set_ylim((axis_extent[2], axis_extent[3]))
    ax.set_axis_off()
    nx.draw_networkx_nodes(net,
                     pos=xycoords,
                     node_color='k',
                     node_size=5,
                     alpha=0.2)
    if edges is None:
        edges = net.edges()
    
    if linewidths is None:
        linewidths = np.ones(len(edges))

    for edge, lw  in zip(edges, linewidths):
        nx.draw_networkx_edges(
            net,
            pos=xycoords,
            edge_color='r',
            width=lw,
            edgelist=[edge],
            alpha=lw,
        )
    return fig, ax

class SI_AnimHelper(object):
    """
    A class used for creating animated visualizations of the epidemic. 
    For a function making use of this class, see visualize_si below.
    """
    def __init__(self, infection_times, event_fname, airport_info_csv_name, bg_figname):
        if not os.path.exists(event_fname):
            raise IOError("File " + event_fname + "could not be found")

        self.ed = np.genfromtxt(
            event_fname,
            delimiter=' ',
            dtype=None,
            names=True
        )

        self.infection_times = infection_times

        if not os.path.exists(airport_info_csv_name):
            raise IOError("File " + event_fname + "could not be found")

        id_data = np.genfromtxt(
            airport_info_csv_name,
            delimiter=',',
            dtype=None,
            names=True
        )
        self.xcoords = id_data['xcoordviz']
        self.ycoords = id_data['ycoordviz']
        self.fig = plt.figure(figsize=(10, 10))
        self.ax = self.fig.add_axes([0, 0, 1, 1])
        img = plt.imread(bg_figname)
        self.axis_extent = (-6674391.856090588, 4922626.076444283,
                            -2028869.260519173, 4658558.416671531)
        self.img = self.ax.imshow(img, extent=self.axis_extent)
        self.ax.set_xlim((self.axis_extent[0], self.axis_extent[1]))
        self.ax.set_ylim((self.axis_extent[2], self.axis_extent[3]))
        self.ax.set_axis_off()
        self.time_text = self.ax.text(
            0.1, 0.1, "", transform=self.ax.transAxes)
        self.scat_planes = self.ax.scatter([], [], s=0.2, color="k")

        self.n = len(self.xcoords)
        self.airport_colors = np.array(
            [[0, 1, 0] for i in range(self.n)], dtype=float)
        self.scat_airports = self.ax.scatter(
            self.xcoords, self.ycoords, c=self.airport_colors, s=5, alpha=0.2)

    def draw(self, frame_time):
        """
        Draws the current situation of the epidemic spreading.

        Parameters
        ----------
        frame_time : int
            the current time in seconds
            (should lie somewhere between 1229231100 and 1230128400)
        """
        time_str = time.asctime(time.gmtime(frame_time))
        self.time_text.set_text(time_str)
        # this should be improved
        unfinished_events = (self.ed['EndTime'] > frame_time)
        started_events = (self.ed['StartTime'] < frame_time)

        # oge = 'on going events'
        oge = self.ed[started_events * unfinished_events]
        fracs_passed = (float(frame_time) - oge['StartTime']) / oge['Duration']
        ongoing_xcoords = ((1 - fracs_passed) * self.xcoords[oge['Source']]
                           + fracs_passed * self.xcoords[oge['Destination']])
        ongoing_ycoords = ((1 - fracs_passed) * self.ycoords[oge['Source']]
                           + fracs_passed * self.ycoords[oge['Destination']])
        self.scat_planes.set_offsets(
            np.array([ongoing_xcoords, ongoing_ycoords]).T)
        self.event_durations = self.ed['Duration']

        infected = (self.infection_times < frame_time)

        self.airport_colors[infected] = (1, 0, 0)  # red
        self.airport_colors[~infected] = (0, 1, 1)  # green
        self.scat_airports = self.ax.scatter(
            self.xcoords, self.ycoords, c=self.airport_colors, s=20, alpha=0.5)

    def draw_anim(self, frame_time):
        self.draw(frame_time)
        return self.time_text, self.scat_planes, self.scat_airports

    def init(self):
        return self.time_text, self.scat_planes, self.scat_airports
    
def visualize_si(infection_times,
                 event_fname,
                 airport_info_csv_name,
                 bg_figname,
                 viz_start_time=1229231100,
                 viz_end_time=1230128400,
                 tot_viz_time_in_seconds=60,
                 fps=10,
                 save_fname=None,
                 writer=None):
    """
    Animates the infection process as a function of time.

    Parameters
    ----------
    infection_times : numpy array
        infection times of the nodes
    event_fname : str
        path of the event file
    airport_info_csv_name : str
        path of the airport info file
    viz_start_time : int
        start time in seconds after epoch
    viz_end_time : int
        end_time in seconds after epoch
    tot_viz_time_in_seconds : int
        length of the animation for the viewer
    fps : int
        frames per second (use low values on slow computers)
    save_fname : str
        where to save the animation
    """
    print(airport_info_csv_name)

    # By default, spans the whole time range of infections.
    times = np.linspace(
        viz_start_time, viz_end_time, fps * tot_viz_time_in_seconds + 1)
    siah = SI_AnimHelper(infection_times,event_fname, airport_info_csv_name, bg_figname)
    ani = FuncAnimation(
        siah.fig, siah.draw_anim, init_func=siah.init, frames=times,
        interval=1000 / fps, blit=True
    )
    if save_fname is not None:
        print("Saving the animation can take quite a while.")

        print("Be patient...")

        ani.save(save_fname, writer='ffmpeg', codec='mpeg4')
    else:
        plt.show()
        
# Befin of the exercise

# Functions used for solving the problems:

# Task 1: Basic implementation

def get_net_and_id_data(network_file_path,airport_info_file_path):
    """
    Reads the network and airport information.
    
    Parameters
    ----------
    network_file_path : str
            path of the network file
    airport_info_file_path : str
            path of the airport info csv file
            
    Returns
    -------
    net : nx.Graph
            the weighted network read from the file
    id_data : pd.DataFrame
            information about airport id's, symbols, names, and cities
    """
    # YOUR CODE HERE
    #TODO: implement a function for reading the network and airport info
    # Hint: make use of pd_read_csv for reading the airport info
    return net, id_data

def simulate_si(infection_prob, event_file_path, airport_info_file_path,
                seed=None, immunized_nodes=[], return_trans_nodes=False):
    """
    General function for simulating the SI dynamics.

    Parameters
    ----------
    infection_prob : float
        the probability of infection per flight
    event_file_path : str
        path of the event file
    airport_info_file_path : str
        path of the airport info file
    seed : int
        the seed node index
    immunized_nodes : list
        indexes of nodes that have been immunized (used in task 5)
    return_trans_nodes : bool
        if True, the indices of the source nodes of disease-transmitting 
        flights are returned (used in task 6)
        
    Returns:
    --------
    infection_times : list or np.array
        time when each of the nodes gets infected, infection time of 
        non-infected nodes is Inf
    seed : int
        index of the seed node used
    trans_nodes : list or np.array
        the origin of infection (i.e. the source of the infection-carrying
        flight) for each node
    
    """
    infection_times = []
    trans_nodes = []
    # YOUR CODE HERE
    #TODO: implement a function for running the SI simulation
    # Hints: 
    
    # Use infection_times to keep a record of the infection times of airports (it can be either a list
    # or an np.array). infection_times[i] should be the infection time of node i at all times. Initially, 
    # set the infection times of all nodes to np.float('inf'), excluding the seed node (you can set the 
    # data set’s first flight’s departure time as the infection time for the seed). Then, whenever a node becomes
    # infected, set its time of infection to the arrival time of the flight carrying the infection.
    
    # To implement the model, first sort all flights in increasing order of departure time, and then go through 
    # them in sequence. Is the source node infected at the current time (look this up in infection_times)? If so,
    # check if the infection time of the target node is larger that the arrival time of the flight. If so, the 
    # flight infects the target node with probability p. In this case, set the target node's infection time to
    # the flight's arrival time.
    
    # pd.read_csv is handy for reading the event and airport data.

    # You can use iterrows() from pandas to loop over flights.
    
    if return_trans_nodes:
        return infection_times, seed, trans_nodes
    return infection_times, seed

# Task 2: Effect of infection probability on spreading speed

def get_time_bins(event_file_path, n_time_bins):
    """
    Reads from the event file the flight departure and
    arrival times and divided the time from the first departure
    to the last arrival into n_time_bins slots.
    
    Parameters:
    -----------
    event_file_path : str
        path of the event file
    n_time_bins : int
        number of time bins
        
    Returns:
    --------
    time_bins : np.array
        the time bins used for calculating prevalence
    """
    time_bins = []
    # YOUR CODE HERE
    #TODO: write a function for getting the time bins
    #Hint: make use of np.linspace
    
    return time_bins

def calculate_prevalence(infection_times, time_bins):
    """
    Calculates the prevalence of the infection 
    (i.e. fraction of infected nodes) as a function
    of time. Prevalence is calculated in linear time bins,
    so that prevalence of each bin is the fraction of nodes
    infected before the bin.
    
    Parameters:
    -----------
    infection_times : list or np.array
        time when each of the nodes gets infected, infection time of 
        non-infected nodes is Inf
    n_time_bins : int
        number of time bins
        
    Returns:
    --------
    prevalence : np.array
        prevalence (i.e. fraction of infected nodes) in each time bin
    """
    prevalence = []      
    
    # YOUR CODE HERE
    #TODO: write a function for calculating prevalence as a function of time
    # Hint: to get the prevalence, count the number of nodes infected before each time step step and normalize by 
    # the total number of nodes.
    
    prevalence = np.array(prevalence)
    
    return prevalence

# Task 4: Where to hide?

def dict_to_vec(d):
    """
    Transforms a dictionary (e.g. a dictionary of network properties 
    returned by some networkX function) or a networkX DegreeView object
    to a list sorted in the order of dictionary keys.
    
    Parameters:
    -----------
    d : dictionary or networkX DegreeView object
    
    Returns:
    --------
    vec : list
        list of d.values() sorted by d.keys()
    """
    vec = []
    if isinstance(d, dict):
        for key in sorted(d.keys()):
            vec.append(d[key])
    elif isinstance(d, nx.classes.reportviews.DegreeView):
        nodes = sorted([d1[0] for d1 in d])
        for node in nodes:
            vec.append(d[node])
    return vec

if __name__ == "__main__":
    
    # select from here the tasks to perform
    task1 = False
    task2 = False
    task3 = False
    task4 = False
    task5 = False
    task6 = False
    
    # Task 1:
    if task1:
        p = 1.0
        _, id_data = get_net_and_id_data(network_file_path, airport_info_file_path)

        id_seed = 0
        id_target = 0

        # YOUR CODE HERE
        # TODO: add the seed and target node indices

        inf_times, _ = simulate_si(p, event_file_path, 
                                   airport_info_file_path, seed=id_seed)

        print( "Infected node:\n", id_data.loc[id_seed])

        target_infection_time = int(inf_times[id_target])
        print ("--------------------\n\n Target node: ")
        print (id_data.loc[id_target])
        print ("infected at time " + str(target_infection_time) + "\n")
    
    # Task 2:
    if task2:
        
        n_it = 10 # number of iterations
        infection_probs = [0.01, 0.05, 0.1, 0.5, 1.0] # infection probabilities
        id_seed = 0
        n_time_bins = 1000

        _, id_data = get_net_and_id_data(network_file_path, airport_info_file_path)

        prevalence_data = []

        time_bins = get_time_bins(event_file_path, n_time_bins)

        # YOUR CODE HERE
        #TODO: Calculate the prevalence for each infection probability
        # Hints:

        # Repeate simulate_si n_it times for each probability. At each iteration, calculate the prevalence and 
        # average over iterations.

        # For calculating the average, divide the whole time span (from the first departure time to the last 
        # arrival) into equal-sized steps using the get_time_bins function you implemented above. You want to 
        # use same time steps for all infection probabilities (these will define the x axis of the visualization).

        # For the prevalence calculations, use calculate_prevalence defined above.

        # For the next step (visualization), you need a list (or np.array) called prevalence_data
        # that contains the prevalence as a function of time (either list or np.array) for each infection 
        # probability. Each row of this array corresponds to an infection probability, and the number of columns 
        # should equal to n_time_bins.

        prevalence_data = np.array(prevalence_data)

        colors = ['b','g','r','c','m']
        labels = [str(p) for p in infection_probs]
        plot_prevalence(prevalence_data, time_bins, colors, labels)

    # Task 3:
    if task3:
        
        n_it = 10
        p = 0.1
        seed_nodes = [0, 4, 41, 100, 200]
        n_time_bins = 1000

        _, id_data = get_net_and_id_data(network_file_path, airport_info_file_path)

        prevalence_data = []

        time_bins = get_time_bins(event_file_path, n_time_bins)

        # YOUR CODE HERE
        #TODO: Calculate the prevalence for each seed node
        #Hints: 

        # You can re-use your code from task 2; just loop over seed nodes instead of infection probabilities.

        # For the next step (visualization), you need a list (or np.array) called prevalence_data
        # that contains the prevalence as a function of time (either list or np.array) for each seed node.
        # Each row of this array corresponds to a seed node, and the number of columns should equal to
        # n_time_bins.

        prevalence_data = np.array(prevalence_data)

        colors = ['b','g','r','c','m']
        labels = ['seed ' + str(seed) for seed in seed_nodes]
        plot_prevalence(prevalence_data, time_bins, colors, labels)
    
    # Task 4:
    if task4:
        
        p = 0.5
        n_seeds = 50

        net, id_data = get_net_and_id_data(network_file_path, airport_info_file_path)

        inf_time_data = []

        median_inf_times = [] 
        props = []

        # YOUR CODE HERE
        #TODO: Run si_simulator for 50 random seeds to obtain the median infection time; obtain the listed 
        # network properties
        #Hints: 

        # If a node never gets infected, its infection time remains Inf. You want to ignore these infection 
        # times while calculating the median. To this end, you can first replace Infs with nans and then use 
        # np.nanmedian.

        # Feel free to use any networkX functions for obtaining the network properties. These functions return
        # either dictionaries or networkX DegreeView objects; you can use the dict_to_vec function given above
        # to transform these into ordered lists.

        # For the next step (visualization), you need a list (or n.array) of median infection times 
        # (median_infection_times) and a list where each element is a list containing the values of one 
        # network property for all nodes. These lists need to be in the same order.

        prop_names = ["c","k", "s", "bc"]

        plot_scatters_in_subplots(props, median_inf_times, labels=prop_names)

        rs = []

        # YOUR CODE HERE
        #TODO: calculate the Spearman rank correlation between each network property and median infection times 
        #Hint: scipy.stats.spearmanr

        for prop_name, r in zip(prop_names, rs):
            print('Spearman r between median infection time and ' + prop_name + ': ' + str(r))
    
    # Task 5:
    if task5:
        
        p = 0.5
        n_seeds = 20
        n_immunized = 10
        n_time_bins = 1000

        strategy_names = ["c","k","s","bc","random","soc. net. im."]

        net, id_data = get_net_and_id_data(network_file_path, airport_info_file_path)

        prevalence_data = []

        time_bins = get_time_bins(event_file_path, n_time_bins)

        # YOUR CODE HERE
        #TODO: Calculate the prevalence for each immunization strategy
        #Hints:

        # Modify your implementation of simulate_si from task 1 to allow node immunization. To this end, add a list 
        # (or set) of immunized nodes as an optional input, with an empty list (or set) as the default value. Then, 
        # when about to infect a node, check if the node is in the list/set, and if, do not infect.

        # Before running simulate_si, define the immunized nodes for each of the 6 strategies and the seed nodes. 
        # You want to use the same random seed nodes for each immunization strategy. While selecting the seeds, 
        # ensure that they are not among the immunized nodes.

        # You can re-use your code from task 2, this time looping over immunization strategies.

        # For the next step (visualization), you need a list (or np.array) called prevalence_data
        # that contains the prevalence as a function of time (either list or np.array) for each seed node.
        # Each row of this array corresponds to a seed node, and the number of columns should equal to
        # n_time_bins.

        prevalence = np.array(prevalence_data)

        colors = ['b','g','r','c','m','k']
        plot_prevalence(prevalence_data, time_bins, colors, labels=strategy_names)
    
    # Task 6:
    if task6: 
        
        p = 0.5
        n_seeds = 20

        net, id_data = get_net_and_id_data(network_file_path, airport_info_file_path)

        # YOUR CODE HERE
        #TODO: Calculate the fraction of simulations each link carried the disease and construct the maximal spanning
        # tree of the network.

        #Hints:

        # Modify your implementation of simulate_si from task 1 to record which links carried the disease.

        # For running the simulations, you can again re-use your code from task 2.

        # Feel free to use the networkX function for calculating the maximum spanning tree.

        # For the next step (visualization), you need a list of links of the network (called links), a list of 
        # disease transmission fractions per link (called link_trans_fracs), and the maximum spanning tree 
        # (called mst).

        # coordinates of nodes are needed for visualization
        id_data = np.genfromtxt(airport_info_file_path, delimiter=',', dtype=None, names=True, encoding=None)
        xycoords = {}
        for row in id_data:
            xycoords[int(row['id'])] = (row['xcoordviz'], row['ycoordviz'])

        fig, ax  = plot_network_usa(net, xycoords, background_image_file_path, edges=links, linewidths=link_trans_fracs)
        fig.suptitle('Transmit fractions')

        fig, ax = plot_network_usa(net, xycoords, background_image_file_path, edges=mst.edges())
        fig.suptitle('Maximum spanning tree')

        # YOUR CODE HERE
        #TODO: Obtain the weight and edge betweenness of each link.
        #Hints: 

        # nx.edge_betweenness_centrality

        # For visualization purposes, you may want to normalize weights and edge betweenness centralities by  
        # their maximum values.

        # For the next step (visualization), you need lists or np.arrays of link weights (called weights) and 
        # edge betweenness centralities (called ebs).

        props = [weights, ebs]
        prop_names = ['weight', 'edge betweenness']

        plot_scatters_in_subplots(props, link_trans_fracs, labels=prop_names, n_columns=2, \
                                  y_label='link transition fraction', x_labels=[])

        rs = []

        # YOUR CODE HERE
        #TODO: Calculate the Spearman rank correlation between the link-wise properties and link transition 
        # fraction.

        for prop_name, r in zip(prop_names, rs):
            print('Spearman r between median infection time and ' + prop_name + ': ' + str(r))
